@@ -18,7 +18,6 @@ typedef enum {
 
 xcb_pixmap_t draw_image(uint32_t* resolution);
 void redraw_screen(void);
-void start_clear_indicator_timeout(void);
-void stop_clear_indicator_timeout(void);
+void clear_indicator(void);
 
 #endif
